/*****************************************************************************
 * Joseph Zambreno
 * Department of Electrical and Computer Engineering
 * Iowa State University
 *****************************************************************************/

/*****************************************************************************
 * bmp_utils.c - utilities for reading .bmp files. 
 *
 *
 * NOTES:
 * 3/22/11 by JAZ::Design created.
 *****************************************************************************/

#include "bmp_utils.h"


/*****************************************************************************
 * Function: bmpRead                            
 * Description: Converts a bmp file to an equivalent 1D array of uint32_t_
 * values representing pixels.
 *****************************************************************************/
uint8_t *bmpRead(uint8_t *bmp_fname, uint32_t *width, uint32_t *height) {

  uint32_t twidth, theight, rowbytes, i, j;
  uint8_t extra_pixels[4];
  uint8_t *tpixels;
  uint8_t *pixels;
  FILE *infile;
  bmp_file_info *bmp;

  // Open up the file, and read the header information
  infile = fopen(bmp_fname, "rb"); 
  if (!infile) {
    raise_error(ERR_NOFILE1);
  }

  // Allocate memory for the bmp_file_info variable
  bmp = (bmp_file_info *)malloc(sizeof(bmp_file_info));
  if (!bmp) {
    raise_error(ERR_NOMEM);
  }

  bmp->h1 = (bmp_magic *)malloc(sizeof(bmp_magic));
  bmp->h2 = (bmp_header *)malloc(sizeof(bmp_header));
  bmp->h3 = (dib_header *)malloc(sizeof(dib_header));
  if ((!bmp->h1) || (!bmp->h2) || (!bmp->h3)) {
    raise_error(ERR_NOMEM);
  }


  if (config.debug_level > 3) {
    fprintf(stderr, "Reading bmp header information from %s\n", bmp_fname);
  }
    
	
  fread(bmp->h1, sizeof(bmp_magic), 1, infile);
  fread(bmp->h2, sizeof(bmp_header), 1, infile);
  fread(bmp->h3, sizeof(dib_header), 1, infile);

  if (config.debug_level > 3) {
    bmpInfo(bmp);
  }


  // If the file header is more recent than BMPv3, we have to read the rest 
  // of the data (and can then safely skip it)
  if (bmp->h3->header_sz > sizeof(dib_header)) {
    bmp->h4 = (uint8_t *)malloc(bmp->h3->header_sz-sizeof(dib_header));
    if (!bmp->h4) {
      raise_error(ERR_NOMEM);
    }
    fread(bmp->h4, 1, bmp->h3->header_sz-sizeof(dib_header), infile);
  }


  twidth = bmp->h3->width;
  theight = bmp->h3->height;
 
  // Allocate memory for the 1D array
  pixels = (uint8_t *)malloc(twidth * theight * (bmp->h3->bitspp / 8.0));
  if (!pixels) {
    raise_error(ERR_NOMEM);
  }

 
  // Allocate memory for each temporary row from the file
  rowbytes = (uint32_t)ceil(twidth * bmp->h3->bitspp / 8.0);
  tpixels = (uint8_t *)malloc(rowbytes*sizeof(uint8_t));
  if (!tpixels) {
    raise_error(ERR_NOMEM);
  }

  // Read from the file and copy things over. Note that
  // each row must have land on a 4-byte boundary. Read
  // (and ignore) those bytes as necessary. 
  for (i = 0; i < theight; i++) {
    fread(tpixels, 1, rowbytes, infile);

    // The storage in pixels[] is dependent on the bitspp
    // BMP files are in reverse raster scan order (which is what we want for
    // texture mapping anyway)
    switch (bmp->h3->bitspp) {
      case 1:
	for (j = 0; j < rowbytes; j++) {
	  pixels[i*rowbytes+j] = (tpixels[j/8] >> (7-(j%8))) & 0x1;
	}
	break;
      case 4:
	for (j = 0; j < rowbytes; j++) {
	  pixels[i*rowbytes+j] = (tpixels[j/2] >> (16*((1-j%2)))) & 0xF;
	}
	break;
      case 8:
	for (j = 0; j < rowbytes; j++) {
	  pixels[i*rowbytes+j]   = tpixels[j];
	}
	break;
      case 16:
	for (j = 0; j < rowbytes; j+=2) {
	  pixels[i*rowbytes+j]   = tpixels[j+1];
	  pixels[i*rowbytes+j+1] = tpixels[j];
	}
	break;
      case 24:
	for (j = 0; j < rowbytes; j+=3) {
	  pixels[i*rowbytes+j+0] = tpixels[j+2];
	  pixels[i*rowbytes+j+1] = tpixels[j+1];
	  pixels[i*rowbytes+j+2] = tpixels[j+0];
	}
	break;
      case 32:
	for (j = 0; j < rowbytes; j+=4) {
	  pixels[i*rowbytes+j+2] = tpixels[j+0];
	  pixels[i*rowbytes+j+1] = tpixels[j+1];
	  pixels[i*rowbytes+j+0] = tpixels[j+2];
	  pixels[i*rowbytes+j+3] = tpixels[j+3];
	}
	break;
      default:
	break;
    }

    if (rowbytes % 4 != 0) {
      fread(extra_pixels, 1, 4-(rowbytes%4), infile);
    }

  }
     
  fclose(infile);

  *width = twidth;
  *height = theight;
 
  return pixels;
}


/*****************************************************************************
 * Function: bmpInfo                            
 * Description: Print out some statistics about the bmp file, for debug
 * purposes
 *****************************************************************************/
void bmpInfo(bmp_file_info *bmp) {

  printf("Printing bmp file information:\n");
  printf("   Magic number is %x%x\n", bmp->h1->magic[1], bmp->h1->magic[0]);
  printf("   File size is %d\n", bmp->h2->filesz);
  printf("   Header size is %d\n", bmp->h3->header_sz);
  printf("   Color depth is %d\n", bmp->h3->bitspp);
  printf("   Width = %d, Height = %d\n", bmp->h3->width, bmp->h3->height);
  printf("   Number of pixel bytes = %d\n", bmp->h3->bmp_bytesz);

  return;
}



