/*****************************************************************************
 * Joseph Zambreno               
 * Department of Electrical and Computer Engineering
 * Iowa State University
 *****************************************************************************/


/*****************************************************************************
 * boomtown.c - boomtown application. Takes as input several move files and
 * an optional map file and runs a BomberMan-style simulation for up to 
 * four teams.
 *
 *
 * NOTES:
 * 4/15/11 by JAZ::Design created.
 *****************************************************************************/

#include "boomtown.h"
#include "bmp_utils.h"


int main(int argc, char **argv) {

  // Parse the command-line to set config state
  parseArgs(argc, argv);

  if (config.debug_level > 5) {
    printConfig();
  }

  // Initialize and compile the game files to make sure there are no errors
  compileMap();
  initGame();
  compileTeams();


  // Start up the OpenGL library and load the game textures
  initGL(argc, argv);
  loadSprites();


  glutMainLoop();


  return 0;
}




/*****************************************************************************
 * Function: updateGame                            
 * Description: Main update function. Keeps track of the player commands
 *****************************************************************************/
void updateGame() {

  int32_t i, j, k, i2, j2, k2;
  uint8_t cmd_done;
  team_type *team, *team2;
  player_type *player, *player2;
  cmd_type *cmd;

  if (game.done == 1)
    return;


  for (i = 0; i < 4; i++) {

    team = &game.teams[i];

    // Has anyone been hit? If so, start the fall sequence and don't let them
    // execute any more commands
    for (j = 0; j < 4; j++) {
      player = &team->players[j];  
      for (k = 0; k < player->max_bombs; k++) {
	if (player->bombs[k].status == EXPLODE) {
	  for (i2 = 0; i2 < 4; i2++) {
	    team2 = &game.teams[i2];
	    for (j2 = 0; j2 < 4; j2++) {
	      player2 = &team2->players[j2];
	      if ((player2->status == ALIVE) && 
		  (boomIntersect(&player->bombs[k], 
				 player2->x_pos, player2->y_pos))) {
		if (config.debug_level > 1) {
		  printf("Bomb from player %d.%d hit player %d.%d\n",i+1,
			 j+1, i2+1, j2+1);
		}
		// Safeties can't be hit by own team's bombs
		if ((i == i2) && (player2->type == SAFETY)) {
		  if (config.debug_level > 1) {
		    printf("Player %d.%d survives!\n", i2+1, j2+1);
		  }
		}
		else {
		  player2->just_hit++;
		  player2->count = 0;
		  // Update scoreboard (-1 for team hit, +1 for hit)
		  if (i == i2) {
		    game.score[i]--;
		  }
		  else {
		    game.score[i]++;
		  }

		}

	      }
	      for (k2 = 0; k2 < player->max_bombs; k2++) {
		// Compute intersection with each bomb as well
		if (player2->bombs[k2].status == ACTIVE) {
		  if (boomIntersect(&player->bombs[k], 
				    player2->bombs[k2].x_pos, 
				    player2->bombs[k2].y_pos)) {
		    player2->bombs[k2].just_hit = 1;
		    if (config.debug_level > 2) {
		      printf("Bomb from player %d.%d hit bomb from player", 
			     i+1, j+1);
		      printf(" %d.%d\n", i2+1, j2+1);
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }


  // Update the recently-hit players
  for (i = 0; i < 4; i++) {
    team = &game.teams[i];
    for (j = 0; j < 4; j++) {
      player = &team->players[j];  
      if ((player->movetype >= DEAD1) && (player->movetype <= DEAD4)) {
	player->count++;
	if (player->count > DEATH_TIMER) {
	  player->count = 0;
	  player->movetype++;
	}
      }

      player->lives -= player->just_hit;
      // Update only if just defeated
      if ((player->lives  <= 0) && (player->just_hit > 0)) {
	player->movetype = DEAD1;
	player->status = DEAD;
	player->count = 0;
	player->cur_cmd = player->num_cmds;
      }

      player->just_hit = 0;

    }

  }


  for (i = 0; i < 4; i++) {

    team = &game.teams[i];

    for (j = 0; j < 4; j++) {
      player = &team->players[j];    

      // Update the boombs next, since the bomb update can trigger a new boom
      for (k = 0; k < player->max_bombs; k++) {
	if (player->bombs[k].status == EXPLODE) {

	  // Grow the boom using .count
	  if (player->bombs[k].rev == 0) {
	    player->bombs[k].count++;
	    if (player->bombs[k].count == BOOM_COUNT_LIMIT) {
	      player->bombs[k].count = 0;;
	      player->bombs[k].boom++;

	      if (player->bombs[k].boom == NONE3) {
		player->bombs[k].boom = BOOM5;
		player->bombs[k].timer = 0;
		player->bombs[k].rev = 1;
	      }

	    }
	  }
	  // Hold the boom using .timer
	  else if (player->bombs[k].rev == 1) {
	    player->bombs[k].timer++;
	    if (player->bombs[k].timer == BOOM_TIMER) {
	      player->bombs[k].timer = 0;
	      player->bombs[k].count = BOOM_COUNT_LIMIT;
	      player->bombs[k].rev = 2;
	    }
	  }

	  // Shrink the boom using .count
	  else if (player->bombs[k].rev == 2){
	    player->bombs[k].count--;
	    if (player->bombs[k].boom == BOOM1) {
	      player->bombs[k].count = 0;
	      player->bombs[k].status = INACTIVE;
	      player->bombs[k].type = BOMB1;
	      player->cur_bombs++;	      
	    }

	    else if (player->bombs[k].count == 0) {
	      player->bombs[k].count = BOOM_COUNT_LIMIT;
	      player->bombs[k].boom--;
	    }
	  }
	}
      }
    }
  }


  for (i = 0; i < 4; i++) {
    team = &game.teams[i];

    // Update the bombs second, since the player update can trigger a new bomb
    for (j = 0; j < 4; j++) {

      player = &team->players[j];    

      for (k = 0; k < player->max_bombs; k++) {
	if (player->bombs[k].status == ACTIVE) {
	  player->bombs[k].count++;
	  if (player->bombs[k].count == BOMB_COUNT_LIMIT) {
	    player->bombs[k].count = 0;
	    player->bombs[k].type++;
	    if (player->bombs[k].type == NONE2) {
	      player->bombs[k].type = BOMB1;
	      player->bombs[k].timer++;
	      if (player->bombs[k].timer == BOMB_TIMER) {
		player->bombs[k].timer = 0;
		player->bombs[k].radius = player->radius;
		player->bombs[k].status = EXPLODE;
		player->bombs[k].boom = BOOM1;
		player->bombs[k].rev = 0;
	      }
	    }	    
	  }
	}
	if (player->bombs[k].just_hit == 1) {
	  player->bombs[k].just_hit = 0;
	  player->bombs[k].count = 0;
	  player->bombs[k].timer = 0;
	  player->bombs[k].radius = player->radius;
	  player->bombs[k].status = EXPLODE;
	  player->bombs[k].boom = BOOM1;
	  player->bombs[k].rev = 0;
	}
      }
    }
  }


  for (i = 0; i < 4; i++) {
    team = &game.teams[i];
  
    // Grab the next command if the team is still alive
    if (team->status == DEAD) {
      continue;
    }

    for (j = 0; j < 4; j++) {

      player = &team->players[j];
      if ((player->status == DEAD) || (player->cur_cmd == player->num_cmds)) {
	continue;
      }


      cmd_done = 0;
      while (!cmd_done) {

	cmd = &player->cmds[player->cur_cmd];
	switch (cmd->cmd) {
          case LEFT:
	    switch (player->movetype) {
	      case LEFT1:
	      case LEFT2:
	      case LEFT3:
		player->movetype = DOWN1;
		break;
	      case RIGHT1:
	      case RIGHT2:
	      case RIGHT3:
		player->movetype = UP1;
		break;
	      case UP1:
	      case UP2:
	      case UP3:
		player->movetype = LEFT1;
		break;
	      case DOWN1:
	      case DOWN2:
	      case DOWN3:
		player->movetype = RIGHT1;
		break;		
	    }
	    player->cur_cmd++;
	    if (player->cur_cmd == player->num_cmds) {
	      cmd_done = 1;
	    }
  	    break;
          case RIGHT:
	    switch (player->movetype) {
	      case LEFT1:
	      case LEFT2:
	      case LEFT3:
		player->movetype = UP1;
		break;
	      case RIGHT1:
	      case RIGHT2:
	      case RIGHT3:
		player->movetype = DOWN1;
		break;
	      case UP1:
	      case UP2:
	      case UP3:
		player->movetype = RIGHT1;
		break;
	      case DOWN1:
	      case DOWN2:
	      case DOWN3:
		player->movetype = LEFT1;
		break;		
	    }
	    player->cur_cmd++;
	    if (player->cur_cmd == player->num_cmds) {
	      cmd_done = 1;
	    }
	    break;
          case REVERSE:
	    switch (player->movetype) {
	      case LEFT1:
	      case LEFT2:
	      case LEFT3:
		player->movetype = RIGHT1;
		break;
	      case RIGHT1:
	      case RIGHT2:
	      case RIGHT3:
		player->movetype = LEFT1;
		break;
	      case UP1:
	      case UP2:
	      case UP3:
		player->movetype = DOWN1;
		break;
	      case DOWN1:
	      case DOWN2:
	      case DOWN3:
		player->movetype = UP1;
		break;		
	    }
	    player->cur_cmd++;
	    if (player->cur_cmd == player->num_cmds) {
	      cmd_done = 1;
	    }
	    break;
          case MOVE:
	    switch (player->movetype) {
	      case LEFT1:
	      case LEFT2:
	      case LEFT3:
		// If we hit a gridline, flush to the grid and check for 
		// intersection
		if (fabs(player->x_pos - rintf(player->x_pos)) < ERR_TOL) {
		  player->x_pos = rintf(player->x_pos);
		  player->y_pos = rintf(player->y_pos);
		  if (!computeIntersect(player)) {
		    player->x_pos -= 1.0/player->speed;
		  }
		}
		else {
		  player->x_pos -= 1.0/player->speed;
		}

		cmd->var -= 1.0/player->speed;
		if (fabs(cmd->var) < ERR_TOL) {
		  player->cur_cmd++;
		}
		player->count++;
		if (player->count > COUNT_LIMIT) {
		  player->count = 0;
		  player->movetype++;
		  if (player->movetype > LEFT3) {
		    player->movetype = LEFT1;
		  }
		}
		break;
	      case RIGHT1:
	      case RIGHT2:
	      case RIGHT3:
		if (fabs(player->x_pos - rintf(player->x_pos)) < ERR_TOL) {
		  player->x_pos = rintf(player->x_pos);
		  player->y_pos = rintf(player->y_pos);
		  if (!computeIntersect(player)) {
		    player->x_pos += 1.0/player->speed;
		  }
		}
		else {
		  player->x_pos += 1.0/player->speed;
		}

		cmd->var -= 1.0/player->speed;
		if (fabs(cmd->var) < ERR_TOL) {
		  player->cur_cmd++;
		  player->count = 0;
		}
		player->count++;
		if (player->count > COUNT_LIMIT) {
		  player->count = 0;
		  player->movetype++;
		  if (player->movetype > RIGHT3) {
		    player->movetype = RIGHT1;
		  }
		}
		break;
	      case UP1:
	      case UP2:
	      case UP3:
		if (fabs(player->y_pos - rintf(player->y_pos)) < ERR_TOL) {
		  player->y_pos = rintf(player->y_pos);
		  player->x_pos = rintf(player->x_pos);
		  if (!computeIntersect(player)) {
		    player->y_pos -= 1.0/player->speed;
		  }
		}
		else {
		  player->y_pos -= 1.0/player->speed;
		}

		cmd->var -= 1.0/player->speed;
		if (fabs(cmd->var) < ERR_TOL) {
		  player->cur_cmd++;
		}
		player->count++;
		if (player->count > COUNT_LIMIT) {
		  player->count = 0;
		  player->movetype++;
		  if (player->movetype > UP3) {
		    player->movetype = UP1;
		  }
		}
		break;
	      case DOWN1:
	      case DOWN2:
	      case DOWN3:
		if (fabs(player->y_pos - rintf(player->y_pos)) < ERR_TOL) {
		  player->y_pos = rintf(player->y_pos);
		  player->x_pos = rintf(player->x_pos);
		  if (!computeIntersect(player)) {
		    player->y_pos += 1.0/player->speed;
		  }
		}
		else {
		  player->y_pos += 1.0/player->speed;
		}


		cmd->var -= 1.0/player->speed;
		if (fabs(cmd->var) < ERR_TOL) {
		  player->cur_cmd++;
		}
		player->count++;
		if (player->count > COUNT_LIMIT) {
		  player->count = 0;
		  player->movetype++;
		  if (player->movetype > DOWN3) {
		    player->movetype = DOWN1;
		  }
		}
		break;
	    }
	    cmd_done = 1;
	    break;
	
        case DROP:

	  if (player->cur_bombs > 0) {

	    // For now, we flush the player to the grid before dropping
	    player->x_pos = rintf(player->x_pos);
	    player->y_pos = rintf(player->y_pos);

	    for (k = 0; k < player->max_bombs; k++) {
	      if (player->bombs[k].status == INACTIVE) {
		player->bombs[k].status = ACTIVE;
		player->bombs[k].timer = 0;
		player->bombs[k].count = 0;
		player->bombs[k].x_pos = player->x_pos;
		player->bombs[k].y_pos = player->y_pos;
		player->bombs[k].type = BOMB1;
		player->bombs[k].boom = NONE3;
		player->cur_bombs--;
		break;
	      }
	    }
	    player->cur_cmd++;
	  }
	  // We end the command even if the bomb isn't dropped. Becomes a 
	  // defacto WAIT. 
	  cmd_done = 1;
	  break;
        case WAIT:

	  cmd->var -= 1.0/player->speed;
	  if (fabs(cmd->var) < ERR_TOL) {
	    // For now, we flush the player to the grid before moving on
	    player->x_pos = rintf(player->x_pos);
	    player->y_pos = rintf(player->y_pos);
	    player->cur_cmd++;
	  }
	  cmd_done = 1;
	  break;	
	}
      }

    }

  }

}



/*****************************************************************************
 * Function: updateDone                            
 * Description: Are we done yet? And if so, display a finishing sequence
 *****************************************************************************/
void updateDone() {

  int32_t i, j, k;
  int8_t max_score;

  if (game.done == 1) {

    // If a player is still sitting around (dead), end their misery
    for (i = 0; i < 4; i++) {
      for (j = 0; j < 4; j++) {
	if (game.teams[i].players[j].movetype >= DEAD1 && 
	    game.teams[i].players[j].movetype <= DEAD4) {
	  game.teams[i].players[j].movetype = NONE1;
	}
      }
    }


    for (i = 0; i < 4; i++) {
      if (game.teams[i].winner == 1) {

	for (j = 0; j < 4; j++) {
	  if (game.teams[i].players[j].status == ALIVE) {
	    game.teams[i].players[j].count++;
	    if (game.teams[i].players[j].count == 3*COUNT_LIMIT) {
	      game.teams[i].players[j].movetype++;
	      game.teams[i].players[j].count = 0;
	      if (game.teams[i].players[j].movetype == DOWN3) {
		game.teams[i].players[j].movetype = DOWN1;
	      }
	      else {
		game.teams[i].players[j].movetype == DOWN3;
	      }
	    }
	  }
	}

      }

    }

  }

  // We're not done if there are still commands or bombs available
  else {
    game.done = 1;
    for (i = 0; i < 4; i++) {
      for (j = 0; j < 4; j++) {
	if (game.teams[i].players[j].cur_cmd < 
	    game.teams[i].players[j].num_cmds) {
	  // If the remaining commands are all WAITs, we can quit
	  for (k = game.teams[i].players[j].cur_cmd; 
	       k < game.teams[i].players[j].num_cmds; k++) {
	    if (game.teams[i].players[j].cmds[k].cmd != WAIT) {
	      game.done = 0;
	      break;
	    }
	  }
	}
	for (k = 0; k < game.teams[i].players[j].max_bombs; k++) {
	  if (game.teams[i].players[j].bombs[k].status != INACTIVE) {
	    game.done = 0;
	    break;
	  }
	}

      }

    }

    // If the 'e' key has been pressed, we will end early anyway. 
    if (game.finish == 1) {
      game.done = 1;
    }

    // Calculate who has won
    if (game.done == 1) {
      max_score = -5;
      printf("\n----------------------------------------------\n");
      printf("Game over - Total scores:\n");
      for (i = 0; i < 4; i++) {
	if (game.score[i] > max_score) {
	  max_score = game.score[i];
	}
	printf("Team %8s scored %i\n", game.teams[i].name, game.score[i]);
      }
      printf("----------------------------------------------\n\n");
      for (i = 0; i < 4; i++) {
	if (game.score[i] == max_score) {
	  game.teams[i].winner = 1;
	  printf("Team %8s wins!\n", game.teams[i].name);
	}
      }

      // If we've calculated that a given team has won, do the first step above
      for (i = 0; i < 4; i++) {
	if (game.teams[i].winner == 1) {

	  for (j = 0; j < 4; j++) {
	    if (game.teams[i].players[j].status == ALIVE) {
	      game.teams[i].players[j].count = 0;
	      game.teams[i].players[j].movetype = DOWN1;
	    }
	  }
	}
      }
    }
  }

}


/*****************************************************************************
 * Function: displayGame                            
 * Description: Main display function. Draws the field, players, and bombs. 
 *****************************************************************************/
void displayGame() {

  int32_t i, j, k;

  timenew = glutGet(GLUT_ELAPSED_TIME);
  timediff = timenew - timeold;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  drawMapgrid();

  // Draw the scoreboard
  drawScoreboard();


  // Draw any bombs currently on the map
  glBindTexture(GL_TEXTURE_2D, textures[5]);
  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      for (k = 0; k < MAX_BOMBS; k++) {
	if (game.teams[i].players[j].bombs[k].status == ACTIVE) {
	  drawBomb(&game.teams[i].players[j].bombs[k]);
	}
      }
    }
  }


  // If we have explosions, try to draw them as well too
  glBindTexture(GL_TEXTURE_2D, textures[6]);
  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      for (k = 0; k < MAX_BOMBS; k++) {
	if (game.teams[i].players[j].bombs[k].status == EXPLODE) {
	  drawBoom(&game.teams[i].players[j].bombs[k]);
	}
      }
    }
  }


  for (i = 0; i < 4; i++) {
    glBindTexture(GL_TEXTURE_2D, textures[i+1]);
    for (j = 0; j < 4; j++) {
      drawPlayer(&game.teams[i].players[j]);
    }
  }

  glutSwapBuffers();

  
  if ((game.start == 1) && (game.done == 0)) {
    updateGame();
  }


  updateDone();

  timeold = glutGet(GLUT_ELAPSED_TIME);

}


/*****************************************************************************
 * Function: boomIntersect                            
 * Description: Compute boom-player and boom-bomb intersections. Split out
 * here to help simplify updateGame().
 *****************************************************************************/
uint32_t boomIntersect(bomb_type *bomb, float x_pos, float y_pos) {

  float x_b, y_b;
  float x, y;
  int32_t i;

  x = x_pos;
  y = y_pos;
  x_b = bomb->x_pos;//rintf(bomb->x_pos);
  y_b = bomb->y_pos;//rintf(bomb->y_pos);



  /* Remember that the x/y position specifies the upper-lefthand corner of
   * the intersection area. So for x_pos, if boom is at 3.0, and the player
   * is at 3.0-ERR_TOL, it's an intersection. If the player is at 4.0-ERR_TOL, 
   * it is not (position can be assumed to be rounded up). 
   */

  // If we're standing on the bomb, go boom
  if (intersectEq(x_b, y_b, x, y)) {
    return 1;
  }

  // Check left
  for (i = 0; i < bomb->radius; i++) {
    x_b = x_b - 1;
    if (x_b < 0) {
      break;
    }
    if (game.gridmap[(int32_t)rintf(y_b)][(int32_t)rintf(x_b)] == 'X') {
      break;
    }
    if (intersectEq(x_b, y_b, x, y)) { 
      return 1;
    }   
  }
  x_b = bomb->x_pos;

  // Check right
  for (i = 0; i < bomb->radius; i++) {
    x_b = x_b + 1;
    if (x_b >= (game.grid_width)) {
      break;
    }
    if (game.gridmap[(int32_t)rintf(y_b)][(int32_t)rintf(x_b)] == 'X') {
      break;
    }
    if (intersectEq(x_b, y_b, x, y)) { 
      return 1;
    }   
  }
  x_b = bomb->x_pos;

  // Check Up
  for (i = 0; i < bomb->radius; i++) {
    y_b = y_b - 1;
    if (y_b < 0) {
      break;
    }
    if (game.gridmap[(int32_t)rintf(y_b)][(int32_t)rintf(x_b)] == 'X') {
      break;
    }
    if (intersectEq(x_b, y_b, x, y)) {  
      return 1;
    }   
  }
  y_b = bomb->y_pos;

  // Check Down
  for (i = 0; i < bomb->radius; i++) {
    y_b = y_b + 1;
    if (y_b >= (game.grid_height)) {
      break;
    }
    if (game.gridmap[(int32_t)rintf(y_b)][(int32_t)rintf(x_b)] == 'X') {
      break;
    }
    if (intersectEq(x_b, y_b, x, y)) {  
      return 1;
    }   
  }
  y_b = bomb->y_pos;


  return 0;

}



/*****************************************************************************
 * Function: intersectEq
 * Description: Compute the boom intersection equation. Split out here to 
 * simplify boomIntersect
 *****************************************************************************/
uint32_t intersectEq(float x_b, float y_b, float x, float y) {

  //  if (((fabs(x_b-x) <= ERR_TOL) || (((x-x_b) < (1-ERR_TOL)) && (x > x_b))) &&
  //      ((fabs(y_b-y) <= ERR_TOL) || (((y_b-y) < (1-ERR_TOL)) && (y_b > y)))) {
  if((fabs(x_b-x) < (1-ERR_TOL)) && (fabs(y_b-y) < (1-ERR_TOL))) {
    return 1;
  }  

  return 0;
}


/*****************************************************************************
 * Function: computeIntersect                            
 * Description: Compute player-wall and player-bomb intersections. Split out
 * here to help simplify updateGame().
 *****************************************************************************/
 uint32_t computeIntersect(player_type *player) {

   uint32_t intersect = 0;
   uint8_t i, j, k;


   switch (player->movetype) {

     case LEFT1:
     case LEFT2:
     case LEFT3:
       if ((int32_t)player->x_pos <= 0) {
	 intersect = 1;
	 return intersect;
       }
       if (game.gridmap[(int32_t)player->y_pos][(int32_t)player->x_pos-1] 
	   == 'X') {
	 intersect = 1;
	 return intersect;
       }
       if (player->type == GHOST) {
	 intersect = 0;
	 return intersect;
       }
       for (i = 0; i < 4; i++) {
	 for (j = 0; j < 4; j++) {
	   for (k = 0; k < game.teams[i].players[j].max_bombs; k++) {
	     if (game.teams[i].players[j].bombs[k].status == ACTIVE) {
	       if (((int32_t)game.teams[i].players[j].bombs[k].x_pos ==
		   (int32_t)player->x_pos-1) &&
		   ((int32_t)game.teams[i].players[j].bombs[k].y_pos ==
		    (int32_t)player->y_pos)) {
		 intersect = 1;
		 return intersect;
	       }
	     }
	   }
	 }
       }
       
       break;

     case RIGHT1:
     case RIGHT2:
     case RIGHT3:
       if ((int32_t)player->x_pos >= (game.grid_width-1)) {
	 intersect = 1;
	 return intersect;
       }
       if (game.gridmap[(int32_t)player->y_pos][(int32_t)player->x_pos+1] 
	   == 'X') {
	 intersect = 1;
	 return intersect;
       }
       if (player->type == GHOST) {
	 intersect = 0;
	 return intersect;
       }
       for (i = 0; i < 4; i++) {
	 for (j = 0; j < 4; j++) {
	   for (k = 0; k < game.teams[i].players[j].max_bombs; k++) {
	     if (game.teams[i].players[j].bombs[k].status == ACTIVE) {
	       if (((int32_t)game.teams[i].players[j].bombs[k].x_pos ==
		   (int32_t)player->x_pos+1) &&
		   ((int32_t)game.teams[i].players[j].bombs[k].y_pos ==
		    (int32_t)player->y_pos)) {
		 intersect = 1;
		 return intersect;
	       }
	     }
	   }
	 }
       }
       
       break;

     case UP1:
     case UP2:
     case UP3:
       if ((int32_t)player->y_pos <= 0) {
	 intersect = 1;
	 return intersect;
       }
       if (game.gridmap[(int32_t)player->y_pos-1][(int32_t)player->x_pos] 
	   == 'X') {
	 intersect = 1;
	 return intersect;
       }
       if (player->type == GHOST) {
	 intersect = 0;
	 return intersect;
       }
       for (i = 0; i < 4; i++) {
	 for (j = 0; j < 4; j++) {
	   for (k = 0; k < game.teams[i].players[j].max_bombs; k++) {
	     if (game.teams[i].players[j].bombs[k].status == ACTIVE) {
	       if (((int32_t)game.teams[i].players[j].bombs[k].x_pos ==
		   (int32_t)player->x_pos) &&
		   ((int32_t)game.teams[i].players[j].bombs[k].y_pos ==
		    (int32_t)player->y_pos-1)) {
		 intersect = 1;
		 return intersect;
	       }
	     }
	   }
	 }
       }
       
       break;


     case DOWN1:
     case DOWN2:
     case DOWN3:
       if ((int32_t)player->y_pos >= (game.grid_height-1)) {
	 intersect = 1;
	 return intersect;
       }
       if (game.gridmap[(int32_t)player->y_pos+1][(int32_t)player->x_pos] 
	   == 'X') {
	 intersect = 1;
	 return intersect;
       }
       if (player->type == GHOST) {
	 intersect = 0;
	 return intersect;
       }
       for (i = 0; i < 4; i++) {
	 for (j = 0; j < 4; j++) {
	   for (k = 0; k < game.teams[i].players[j].max_bombs; k++) {
	     if (game.teams[i].players[j].bombs[k].status == ACTIVE) {
	       if (((int32_t)game.teams[i].players[j].bombs[k].x_pos ==
		   (int32_t)player->x_pos) &&
		   ((int32_t)game.teams[i].players[j].bombs[k].y_pos ==
		    (int32_t)player->y_pos+1)) {
		 intersect = 1;
		 return intersect;
	       }
	     }
	   }
	 }
       }
       
       break;



   }

   return intersect;
 }



/*****************************************************************************
 * Function: compileMap                            
 * Description: Compiles the individual .bpl and .map files to initialize
 * the game state.
 *****************************************************************************/
void compileMap() {

  FILE *map_file;
  uint8_t linebuf[256], ntok;
  uint8_t size_flag;
  uint32_t i, j;

  // Compile the map file
  if (config.debug_level > 1) {
    printf("Compiling map file %s\n", config.map_fname);
  }

  map_file = fopen(config.map_fname, "r");
  if (!map_file) {
    raise_error(ERR_NOFILE3);
  }

  size_flag = 0;
  i = 1;
  while (!feof(map_file)) {

    fgets(linebuf, 256, map_file);

    // If we have a '#', the line is a comment and we can skip it
    // We can also skip blank lines
    if ((linebuf != 0) && (linebuf[0] != '#') && 
	(linebuf[0] != 10) && (linebuf[0] != 13)) {      

      // Grab the width and height information
      if (size_flag == 0) {
	ntok = sscanf(linebuf, "%hu, %hu", &game.grid_width, &game.grid_height);
	if (ntok != 2) {
	  printf("Error compiling %s, line %d\n", config.map_fname, i);
	  printf("Invalid grid width and height specification");
	  raise_error(ERR_BADFILE3);
	}

	game.gridmap = (uint8_t **)malloc(game.grid_height*sizeof(uint8_t *));
	for (j = 0; j < game.grid_height; j++) {
	  game.gridmap[j] = (uint8_t *)malloc(game.grid_width*sizeof(uint8_t));
	  if (!game.gridmap[j]) {
	    raise_error(ERR_NOMEM);
	  }
	}

	j = 0;
	size_flag = 1;
      }
      else {
	if (j >= game.grid_height) {
	  printf("Error compiling %s, line %d\n", config.map_fname, i);
	  printf("Expected %d valid rows, found %d\n", game.grid_height, j);
	}

	strncpy(game.gridmap[j], linebuf, game.grid_width);
	j++;
      }
    }

    i++;
  }

  fclose(map_file);
  if (j < game.grid_height) {
    printf("Error compiling %s\n", config.map_fname);
    printf("Only %d valid rows detected (expected %d)\n", j, game.grid_height);
    raise_error(ERR_BADFILE3);
  }

}


/*****************************************************************************
 * Function: compileTeams                            
 * Description: Compiles the individual .bpl files to initialize
 * the game state.
 *****************************************************************************/
void compileTeams() {

  FILE *team_file;
  uint8_t linebuf[256], str[16], ntok, len;
  uint8_t player_flag;
  uint32_t i, j, i_team, i_player, temp, temp2, num_cmds;
  team_type *team;
  player_type *player;


  for (i_team = 0; i_team < 4; i_team++) {

    // Compile the .bpl file
    if (config.debug_level > 1) {
      printf("Compiling team file %s\n", config.team_fname[i_team]);
    }

    team_file = fopen(config.team_fname[i_team], "r");
    if (!team_file) {
      raise_error(ERR_NOFILE2);
    }

    team = &game.teams[i_team];


    // Copy the file name, removing the .bpl extension
    len = strlen(config.team_fname[i_team]);
    team->name = (uint8_t *)malloc(len-3);
    if (!team->name) {
      raise_error(ERR_NOMEM);
    }
    strncpy(team->name, config.team_fname[i_team], len-4);
    team->name[len-4] = 0;


    i = 1;
    player_flag = 0;
    while (!feof(team_file)) {
    
      fgets(linebuf, 256, team_file);
    
      // If we have a '#', the line is a comment and we can skip it
      // We can also skip blank lines
      if ((linebuf[0] != '#') && (linebuf[0] != 10) && (linebuf[0] != 13)) { 


	// Grab the player type information
	if (player_flag < 4) {
	  ntok = sscanf(linebuf, "P%i %s", &temp, str);
	  if ((ntok != 2) || (temp != player_flag+1)) {
	    printf("Error compiling %s, line %d\n",config.team_fname[i_team],i);
	    printf("Invalid player type specification\n");
	    raise_error(ERR_BADFILE2);
	  }
	  if ((!strcmp(str, "Captain")) || (!strcmp(str, "CAPTAIN"))) {
	    team->players[player_flag].type = CAPTAIN;
	    player_flag++;
	  }
	  else if ((!strcmp(str, "Ghost")) || (!strcmp(str, "GHOST"))) {
	    team->players[player_flag].type = GHOST;
	    team->players[player_flag].speed = SLOW_SPEED;
	    team->players[player_flag].radius = DEFAULT_RADIUS;
	    team->players[player_flag].max_bombs = LESS_BOMBS;
	    team->players[player_flag].cur_bombs = LESS_BOMBS;	    
	    player_flag++;
	  }
	  else if ((!strcmp(str, "Stretcher")) || (!strcmp(str, "STRETCHER"))) {
	    team->players[player_flag].type = STRETCHER;
	    team->players[player_flag].speed = SLOW_SPEED;
	    team->players[player_flag].radius = BIG_RADIUS;
	    team->players[player_flag].max_bombs = DEFAULT_BOMBS;
	    team->players[player_flag].cur_bombs = DEFAULT_BOMBS;	    
	    player_flag++;
	  }
	  else if ((!strcmp(str, "Safety")) || (!strcmp(str, "SAFETY"))) {
	    team->players[player_flag].type = SAFETY;
	    team->players[player_flag].speed = DEFAULT_SPEED;
	    team->players[player_flag].radius = SMALL_RADIUS;
	    team->players[player_flag].max_bombs = LESS_BOMBS;
	    team->players[player_flag].cur_bombs = LESS_BOMBS;	    
	    player_flag++;
	  }
	  else if ((!strcmp(str, "Sprinter")) || (!strcmp(str, "SPRINTER"))) {
	    team->players[player_flag].type = SPRINTER;
	    team->players[player_flag].speed = DEFAULT_SPEED;
	    team->players[player_flag].radius = DEFAULT_RADIUS;
	    team->players[player_flag].max_bombs = DEFAULT_BOMBS;
	    team->players[player_flag].cur_bombs = DEFAULT_BOMBS;	    
	    player_flag++;
	  }
	  else if ((!strcmp(str, "Hoarder")) || (!strcmp(str, "HOARDER"))) {
	    team->players[player_flag].type = HOARDER;
	    team->players[player_flag].speed = SLOW_SPEED;
	    team->players[player_flag].radius = DEFAULT_RADIUS;
	    team->players[player_flag].max_bombs = MOST_BOMBS;
	    team->players[player_flag].cur_bombs = MOST_BOMBS;	    
	    player_flag++;
	  }
	  else if ((!strcmp(str, "Hulk")) || (!strcmp(str, "HULK"))) {
	    team->players[player_flag].type = HULK;
	    team->players[player_flag].speed = SLOWEST_SPEED;
	    team->players[player_flag].radius = DEFAULT_RADIUS;
	    team->players[player_flag].lives = MORE_LIVES;
	    team->players[player_flag].max_bombs = MORE_BOMBS;
	    team->players[player_flag].cur_bombs = MORE_BOMBS;	    
	    player_flag++;
	  }
	  else {
	    printf("Error compiling %s, line %d\n",config.team_fname[i_team],i);
	    printf("Invalid player type specification\n");
	    raise_error(ERR_BADFILE2);
	  }

	}

	else {

	  ntok = sscanf(linebuf, "P%i.%s %i", &temp, str, &temp2);
	  if ((ntok != 2) && (ntok != 3)) {
	    printf("Error compiling %s, line %d\n",config.team_fname[i_team],i);
	    printf("Invalid move type specification\n");
	    raise_error(ERR_BADFILE2);
	  }
	  if ((temp < 1) || (temp > 4)) {
	    printf("Error compiling %s, line %d\n",config.team_fname[i_team],i);
	    printf("Team specifier P%d is invalid\n", temp);
	    raise_error(ERR_BADFILE2);
	  }

	  if ((!strcmp(str, "Left")) || (!strcmp(str, "LEFT"))) {
	    if (ntok == 3) {
	      printf("Error compiling %s, line %d\n",config.team_fname[i_team],i);
	      printf("Command \"LEFT\" requires no modifier\n");
	      raise_error(ERR_BADFILE2);
	    }
	    team->players[temp-1].num_cmds++;
	    num_cmds = team->players[temp-1].num_cmds;
	    team->players[temp-1].cmds = (cmd_type *)realloc(
		      team->players[temp-1].cmds, num_cmds*sizeof(cmd_type));
	    team->players[temp-1].cmds[num_cmds-1].cmd = LEFT;
	    team->players[temp-1].cmds[num_cmds-1].var = 0.0;
	  }
	  else if ((!strcmp(str, "Right")) || (!strcmp(str, "RIGHT"))) {
	    if (ntok == 3) {
	      printf("Error compiling %s, line %d\n",config.team_fname[i_team],i);
	      printf("Command \"RIGHT\" requires no modifier\n");
	      raise_error(ERR_BADFILE2);
	    }
	    team->players[temp-1].num_cmds++;
	    num_cmds = team->players[temp-1].num_cmds;
	    team->players[temp-1].cmds = (cmd_type *)realloc(
		      team->players[temp-1].cmds, num_cmds*sizeof(cmd_type));
	    team->players[temp-1].cmds[num_cmds-1].cmd = RIGHT;
	    team->players[temp-1].cmds[num_cmds-1].var = 0.0;
	  }
	  else if ((!strcmp(str, "Reverse")) || (!strcmp(str, "REVERSE"))) {
	    if (ntok == 3) {
	      printf("Error compiling %s, line %d\n",config.team_fname[i_team],i);
	      printf("Command \"REVERSE\" requires no modifier\n");
	      raise_error(ERR_BADFILE2);
	    }
	    team->players[temp-1].num_cmds++;
	    num_cmds = team->players[temp-1].num_cmds;
	    team->players[temp-1].cmds = (cmd_type *)realloc(
		      team->players[temp-1].cmds, num_cmds*sizeof(cmd_type));
	    team->players[temp-1].cmds[num_cmds-1].cmd = REVERSE;
	    team->players[temp-1].cmds[num_cmds-1].var = 0.0;
	  }
	  else if ((!strcmp(str, "Move")) || (!strcmp(str, "MOVE"))) {
	    team->players[temp-1].num_cmds++;
	    num_cmds = team->players[temp-1].num_cmds;
	    team->players[temp-1].cmds = (cmd_type *)realloc(
		      team->players[temp-1].cmds, num_cmds*sizeof(cmd_type));
	    team->players[temp-1].cmds[num_cmds-1].cmd = MOVE;
	    if (ntok == 3) {
	      team->players[temp-1].cmds[num_cmds-1].var = 1.0*temp2;
	    }
	    else {
	      team->players[temp-1].cmds[num_cmds-1].var = 1.0;
	    }
	  }
	  else if ((!strcmp(str, "Drop")) || (!strcmp(str, "DROP"))) {
	    if (ntok == 3) {
	      printf("Error compiling %s, line %d\n", config.team_fname[i_team], i);
	      printf("Command \"DROP\" requires no modifier\n");
	      raise_error(ERR_BADFILE2);
	    }
	    team->players[temp-1].num_cmds++;
	    num_cmds = team->players[temp-1].num_cmds;
	    team->players[temp-1].cmds = (cmd_type *)realloc(
		      team->players[temp-1].cmds, num_cmds*sizeof(cmd_type));
	    team->players[temp-1].cmds[num_cmds-1].cmd = DROP;
	    team->players[temp-1].cmds[num_cmds-1].var = 0.0;
	  }
	  else if ((!strcmp(str, "Wait")) || (!strcmp(str, "WAIT"))) {
	    team->players[temp-1].num_cmds++;
	    num_cmds = team->players[temp-1].num_cmds;
	    team->players[temp-1].cmds = (cmd_type *)realloc(
		      team->players[temp-1].cmds, num_cmds*sizeof(cmd_type));
	    team->players[temp-1].cmds[num_cmds-1].cmd = WAIT;
	    if (ntok == 3) {
	      team->players[temp-1].cmds[num_cmds-1].var = 1.0*temp2;
	    }
	    else {
	      team->players[temp-1].cmds[num_cmds-1].var = 1.0;
	    }
	  }
	  else {
	    printf("Error compiling %s, line %d\n", config.team_fname[i_team], i);
	    printf("Command %s not recognized\n", str);
	    raise_error(ERR_BADFILE2);
	  }


	}

      }

      i++;

    }

    fclose(team_file);

  }
}


/*****************************************************************************
 * Function: initGame                            
 * Description: Initializes the player and game state before compiling
 *****************************************************************************/
void initGame() {

  uint32_t i_team, i_player, j;
  player_type *player;

  // Initialize the teams, players and game state
  game.done = 0;
  game.start = 0;
  game.finish = 0;

  for (i_team = 0; i_team < 4; i_team++) {

    game.teams[i_team].status = ALIVE;
    game.teams[i_team].name = NULL;
    game.score[i_team] = 0;
    game.teams[i_team].winner = 0;

    for (i_player = 0; i_player < 4; i_player++) {
      player = &game.teams[i_team].players[i_player];
      player->status = ALIVE;
      player->just_hit = 0;
      player->speed = DEFAULT_SPEED;
      player->radius = DEFAULT_RADIUS;
      player->max_bombs = DEFAULT_BOMBS;
      player->cur_bombs = DEFAULT_BOMBS;
      player->lives = DEFAULT_LIVES;
      player->type = CAPTAIN;
      player->cur_cmd = 0;
      player->num_cmds = 0;
      player->count = 0;
      player->team = i_team;

      for (j = 0; j < MAX_BOMBS; j++) {
	player->bombs[j].status = INACTIVE;
	player->bombs[j].just_hit = 0;
	player->bombs[j].player = i_player;
	player->bombs[j].team = i_team;
      }

      switch (i_team) {
        case 0:
	  player->movetype = DOWN1;
	  switch(i_player) {
	    case 0:
	      player->x_pos = 0.0;
	      player->y_pos = 0.0;
	      break;
	    case 1:
	      player->x_pos = 0.0;
	      player->y_pos = 1.0;
	      break;
	    case 2:
	      player->x_pos = 1.0;
	      player->y_pos = 0.0;
	      break;
	    case 3:
	      player->x_pos = 1.0;
	      player->y_pos = 1.0;
	      break;
	  }
	  break;
	case 1:
	  player->movetype = RIGHT1;
	  switch(i_player) {
	    case 0:
	      player->x_pos = 0.0;
	      player->y_pos = game.grid_height-1.0;
	      break;
	    case 1:
	      player->x_pos = 1.0;
	      player->y_pos = game.grid_height-1.0;
	      break;
	    case 2:
	      player->x_pos = 0.0;
	      player->y_pos = game.grid_height-2.0;
	      break;
	    case 3:
	      player->x_pos = 1.0;
	      player->y_pos = game.grid_height-2.0;
	      break;
	  }
	  break;
        case 2:
	  player->movetype = LEFT1;
	  switch(i_player) {
	    case 0:
	      player->x_pos = game.grid_width-1.0;
	      player->y_pos = 0.0;
	      break;
	    case 1:
	      player->x_pos = game.grid_width-2.0;
	      player->y_pos = 0.0;
	      break;
	    case 2:
	      player->x_pos = game.grid_width-1.0;
	      player->y_pos = 1.0;
	      break;
	    case 3:
	      player->x_pos = game.grid_width-2.0;
	      player->y_pos = 1.0;
	      break;
	  }
	  break;
        case 3:
	  player->movetype = UP1;
	  switch(i_player) {
	    case 0:
	      player->x_pos = game.grid_width-1.0;
	      player->y_pos = game.grid_height-1.0;
	      break;
	    case 1:
	      player->x_pos = game.grid_width-1.0;
	      player->y_pos = game.grid_height-2.0;
	      break;
	    case 2:
	      player->x_pos = game.grid_width-2.0;
	      player->y_pos = game.grid_height-1.0;
	      break;
	    case 3:
	      player->x_pos = game.grid_width-2.0;
	      player->y_pos = game.grid_height-2.0;
	      break;
	  }
	  break;
      }


    }
  }
}

