/*****************************************************************************
 * Joseph Zambreno               
 * Department of Electrical and Computer Engineering
 * Iowa State University
 *****************************************************************************/


/*****************************************************************************
 * boomtown.h - boomtown application. Takes as input several move files and
 * an optional map file and runs a BomberMan-style simulation for up to 
 * four teams.
 *
 *
 * NOTES:
 * 4/15/11 by JAZ::Design created.
 *****************************************************************************/



#ifndef _BOOMTOWN_H_
#define _BOOMTOWN_H_

#include <GL/glut.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#define EXEC_NAME "boomtown"


#define ERR_USAGE 1
#define ERR_NOFILE1 2
#define ERR_NOFILE2 3
#define ERR_NOFILE3 4
#define ERR_BADFILE1 5
#define ERR_BADFILE2 6
#define ERR_BADFILE3 7
#define ERR_NOMEM 50
#define ERR_UNDEFINED 100

#define FRAME_LATENCY 30
#define SCREEN_WIDTH_DEFAULT 40
#define SCREEN_HEIGHT_DEFAULT 40
#define GRID_WIDTH_DEFAULT 16
#define GRID_HEIGHT_DEFAULT 16
#define DEBUG_DEFAULT 0
#define MAP_FNAME_DEFAULT "default.map"
#define TEAM_FNAME_DEFAULT "default.bpl"

#define NUM_TEXTURES 9
#define ERR_TOL (1.0/(SLOWEST_SPEED+1))
#define COUNT_LIMIT 5
#define DEATH_TIMER 6
#define BOMB_COUNT_LIMIT 15
#define BOMB_TIMER 3
#define BOOM_COUNT_LIMIT 5
#define BOOM_TIMER 5

/* Some enums and macros to simplify drawing functions */
enum PLAYER_MOVETYPES {LEFT1, LEFT2, LEFT3, RIGHT1, RIGHT2, RIGHT3, UP1, UP2, 
		       UP3, DOWN1, DOWN2, DOWN3, KICKLEFT, KICKRIGHT, KICKUP, 
		       KICKDOWN, DEAD1, DEAD2, DEAD3, DEAD4, NONE1};
enum BOMB_TYPES {BOMB1, BOMB2, BOMB3, NONE2};
enum BOOM_TYPES {BOOM1, BOOM2, BOOM3, BOOM4, BOOM5, NONE3};
enum BOOM_PARTS {MID, LPART, LEND, RPART, REND, UPART, UEND, DPART, DEND,NONE4};

/* Per-player enums and macros */
enum CMD_TYPES {LEFT, RIGHT, REVERSE, MOVE, DROP, WAIT};
enum PLAYER_TYPES {CAPTAIN, GHOST, STRETCHER, SAFETY, SPRINTER, HOARDER, HULK};
#define ALIVE 0
#define DEAD 1
#define INACTIVE 0
#define ACTIVE 1
#define EXPLODE 2
#define DEFAULT_SPEED 8
#define FAST_SPEED 5
#define FASTEST_SPEED 4
#define SLOW_SPEED 10
#define SLOWEST_SPEED 16
#define DEFAULT_RADIUS 2
#define SMALL_RADIUS 1
#define BIG_RADIUS 4
#define MAX_RADIUS BIG_RADIUS
#define DEFAULT_BOMBS 2
#define LESS_BOMBS 1
#define MORE_BOMBS 3
#define MOST_BOMBS 4
#define MAX_BOMBS MOST_BOMBS
#define DEFAULT_LIVES 1
#define MORE_LIVES 2


/* Structure to hold configuration information */
typedef struct {
  int32_t debug_level;
  uint16_t screen_width;
  uint16_t screen_height;
  uint8_t *team_fname[4];
  uint8_t *map_fname;
} config_type;


/* Command information */
typedef struct {
  enum CMD_TYPES cmd;
  float var;
} cmd_type;


/* Bomb information */
typedef struct {
  uint8_t status;
  enum BOMB_TYPES type;
  enum BOOM_TYPES boom;
  uint8_t just_hit;
  uint8_t timer;
  uint8_t radius;
  uint8_t count;
  uint8_t rev;
  float x_pos;
  float y_pos;
  uint32_t player;
  uint32_t team;
} bomb_type;



/* Per-player state information */
typedef struct {
  enum PLAYER_TYPES type;
  enum PLAYER_MOVETYPES movetype;
  uint8_t status;
  int8_t just_hit;
  int8_t lives;
  uint8_t speed;
  uint8_t radius;
  uint8_t max_bombs;
  uint8_t cur_bombs;
  float x_pos;
  float y_pos;
  uint32_t count;
  uint32_t cur_cmd;
  uint32_t num_cmds;
  cmd_type *cmds;
  bomb_type bombs[MAX_BOMBS];
  uint8_t team;
} player_type;


/* Per-team state information */
typedef struct {
  uint8_t status;
  uint8_t winner;
  uint8_t *name;
  player_type players[4];
} team_type;


/* Structure to hold game state information */
typedef struct {
  int8_t score[4];
  uint8_t start;
  uint8_t done;
  uint8_t finish;
  uint16_t grid_width;
  uint16_t grid_height;
  uint8_t **gridmap;
  team_type teams[4];
} game_type;


/* Application-level globals */
config_type config;
game_type game;
GLuint textures[NUM_TEXTURES];
int32_t timeold, timenew, timediff;


/* Function prototypes (utils.c) */
void print_help();
void raise_error(uint32_t);
void parseArgs(int argc, char **argv);
void printConfig();
void printStatus();

/* Function prototypes (gl_utils.c) */
void initGL(int argc, char **argv);
static void timerCallback(int value);
void keyPress(uint8_t key, int x, int y);
void loadSprites();
void drawBlock(float x, float y);
void drawGround(float x, float y, uint8_t alt);
void drawPlayer(player_type *player);
void drawBomb(bomb_type *bomb);
void drawBoom(bomb_type *bomb);
void drawBoomPart(float x, float y, enum BOOM_TYPES boom, enum BOOM_PARTS part);
void drawScoreboard();
void drawMapgrid();
void drawText(uint8_t *str, uint8_t len, float x_pos, float y_pos, 
	      float width, float height);

/* Function prototyptes (boomtown.c) */
void initGame();
void compileMap();
void compileTeams();
void displayGame();
void updateGame();
void updateDone();
uint32_t computeIntersect(player_type *player);
uint32_t boomIntersect(bomb_type *bomb, float x_cen, float y_cen);
uint32_t intersectEq(float x_b, float y_b, float x, float y);

#endif /* _BOOMTOWN_H_ */
