/*****************************************************************************
 * Joseph Zambreno               
 * Department of Electrical and Computer Engineering
 * Iowa State University
 *****************************************************************************/

/*****************************************************************************
 * utils.c - utility functions for the boomtown application.
 *
 *
 * NOTES:
 * 4/15/11 by JAZ::Design created.
 *****************************************************************************/

#include "boomtown.h"


/*****************************************************************************
 * Function: printStatus                            
 * Description: Prints out the game's status, in terms of player and bomb
 * locations
 *****************************************************************************/
void printStatus() {

  int32_t i, j, k;
  team_type *team;
  player_type *player;

  printf("\nPrinting status information\n");

  for (i = 0; i < 4; i++) {

    team = &game.teams[i];
    printf("-----------------------------------------------\n");
    printf("- Team %d: %s\n", i+1, team->name);
    printf("-- Score %d\n", game.score[i]);
    
    for (j = 0; j < 4 ; j++) {

      player = &game.teams[i].players[j];
      printf("---- Player %d, type %d, status %d, current command number %d\n",
	     j+1, player->type, player->status, player->cur_cmd+1);
      printf("---- Located at [%f, %f]\n", player->x_pos, player->y_pos);
      printf("---- max_bombs = %d, cur_bombs = %d\n", 
	     player->max_bombs, player->cur_bombs);

      for (k = 0; k < 4 ; k++) {
	if (player->bombs[k].status == ACTIVE) {
	  printf("-------- bomb %d (active), located at [%f, %f]\n", k+1,
		 player->bombs[k].x_pos, player->bombs[k].y_pos);
	}
	if (player->bombs[k].status == EXPLODE) {
	  printf("-------- bomb %d (explode), located at [%f, %f]\n", k+1,
		 player->bombs[k].x_pos, player->bombs[k].y_pos);
	}
      }
    
    }

  }


}

/*****************************************************************************
 * Function: print_help                            
 * Description: Prints out the program help message.
 *****************************************************************************/
void print_help() {
  printf("Usage: %s [options] <file1> [<files> ...]\n\n", EXEC_NAME);
  printf("Main options:\n");
  
  printf("-W <value>, --width <value>        Per-grid pixel width ");
  printf("(default is %d)\n", SCREEN_WIDTH_DEFAULT);

  printf("-H <value>, --height <value>       Per-grid pixel height ");
  printf("(default is %d)\n", SCREEN_HEIGHT_DEFAULT);

  printf("-D <value>, --debug <value>        Debug level ");
  printf("(default is %d)\n", DEBUG_DEFAULT);  

  printf("-M <file>, --map <file>            Map file ");
  printf("(default is %s)\n", MAP_FNAME_DEFAULT);

  printf("-h, --help                         Print this message\n");

  printf("<file1> <file2> <file3> <file4>    Team files ");
  printf("(default is %s)\n", TEAM_FNAME_DEFAULT);

  exit(0);
}



/*****************************************************************************
 * Function: raise_error                                    
 * Description: Prints out an error message determined by   
 * the condition and exits the program.                     
 *****************************************************************************/
void raise_error(uint32_t error_num) {

  fprintf(stderr, "\n");
  switch(error_num) {
    case ERR_USAGE:
      fprintf(stderr, "Usage: %s [-W <n>] [-H <n>] [-D <n>] ", EXEC_NAME);
      fprintf(stderr, "[-M <file>] <file1> [<file2> <file3> <file4>]\n");
      break;
    case ERR_NOFILE1:
      fprintf(stderr, "Error: .bmp texture file not found\n");
      break;
    case ERR_NOFILE2:
      fprintf(stderr, "Error: .bpl player file not found\n");
      break;
    case ERR_NOFILE3:
      fprintf(stderr, "Error: .map map file not found\n");
      break;    
    case ERR_BADFILE1:
      fprintf(stderr, "Error: .bmp texture file is invalid\n");
      break;
    case ERR_BADFILE2:
      fprintf(stderr, "Error: .bpl player file is invalid\n");
      break;
    case ERR_BADFILE3:
      fprintf(stderr, "Error: .map map file is invalid\n");
      break;   
    case ERR_NOMEM:
      fprintf(stderr, "Error: application ran out of memory\n");
      break;
    case ERR_UNDEFINED:
    default:
      fprintf(stderr, "Error: undefined error\n");
      break;
  }
  fprintf(stderr, "%s exited with error code %d\n", EXEC_NAME, error_num); 

#if defined (__WIN32) || defined (__WIN64)
  system("pause");
#endif
  exit(error_num);
}



/*****************************************************************************
 * Function: parseArgs
 * Description: Reads in and parses the input command line in order to set 
 * configuration data. 
 *****************************************************************************/
void parseArgs(int argc, char **argv) {
  int8_t i, nfiles;

  /* We require at least two arguments on the command line */
  if (argc < 2) {
    raise_error(ERR_USAGE);
  }

  /* If the last command-line argument is -h or --help that is ok */
  if (!strncmp(argv[argc-1], "-h", 2) || !strncmp(argv[argc-1], "--help", 6)) {
    print_help();
  }

  /* Otherwise, either the last or the last four arguments should be files */
  if (argv[argc-1][0] == '-') {
    raise_error(ERR_USAGE);
  }
  else {
    if (argc > 2) {
      if ((argv[argc-2][0] == '-') || (argv[argc-3][0] == '-')) {
	nfiles = 1;
      }
      else {
	if (argc < 5) {
	  raise_error(ERR_USAGE);
	}
	else if ((argv[argc-2][0] == '-') || (argv[argc-3][0] == '-') || 
		 (argv[argc-4][0] == '-')) {
	  raise_error(ERR_USAGE);
	}
	else {
	  nfiles = 4;
	}
      }
    }
    else {
      nfiles = 1;
    }
  }


  /* Set the default values */
  config.debug_level = DEBUG_DEFAULT;
  config.screen_height = SCREEN_HEIGHT_DEFAULT;
  config.screen_width = SCREEN_WIDTH_DEFAULT;
  config.map_fname = (uint8_t *)malloc(strlen(MAP_FNAME_DEFAULT)+1);
  if (!config.map_fname) {
    raise_error(ERR_NOMEM);
  }
  strcpy(config.map_fname, MAP_FNAME_DEFAULT);  
  for (i = 0; i < 4; i++) {
      config.team_fname[i] = (uint8_t *)malloc(strlen(TEAM_FNAME_DEFAULT)+1);
      if (!config.team_fname[i]) {
	raise_error(ERR_NOMEM);
      }
      strcpy(config.team_fname[i], TEAM_FNAME_DEFAULT);  
  }

  for (i = 1; i < argc-nfiles; i++) {
    /* Arguments must begin with '-' or '--' */
    if (argv[i][0] != '-' || strlen(argv[i]) < 2) {
      raise_error(ERR_USAGE);
    }
   
    /* Decode the single dash arguments */
    if (argv[i][1] != '-') {
      /* Single dash arguments should have just one character after the dash */
      if (strlen(argv[i]) > 2) {
	raise_error(ERR_USAGE);
      }
      switch (argv[i][1]) {
      case 'H':
	i++;
	sscanf(argv[i], "%hu", &config.screen_height);
	break;
      case 'W':
	i++;
	sscanf(argv[i], "%hu", &config.screen_width);
	break;
      case 'D':
	i++;
	sscanf(argv[i], "%d", &config.debug_level);
	break;
      case 'M':
	i++;
	config.map_fname = (char *)realloc(config.map_fname,
					       strlen(argv[i])+1);
	if (!config.map_fname) {
	  raise_error(ERR_NOMEM);
	}
	strcpy(config.map_fname, argv[i]);
	break;
      case 'h':
	print_help();
	break;
      default:
	raise_error(ERR_USAGE);
	break;
      }
    }
       
    /* Decode the double dash arguments */
    else {
      if (!strncmp(argv[i], "--height", 8)) {
	i++;
	sscanf(argv[i], "%hu", &config.screen_height);
      }
      else if (!strncmp(argv[i], "--width", 7)) {
	i++;
	sscanf(argv[i], "%hu", &config.screen_width);
      }
      else if (!strncmp(argv[i], "--debug", 7)) {
	i++;
	sscanf(argv[i], "%d", &config.debug_level);
      }
      else if (!strncmp(argv[i], "--map", 5)) {
	i++;
	config.map_fname = (char *)realloc(config.map_fname,
					   strlen(argv[i])+1);
	if (!config.map_fname) {
	  raise_error(ERR_NOMEM);
	}
	strcpy(config.map_fname, argv[i]);
      }
      else if (!strncmp(argv[i], "--help", 6)) {
	print_help();
      }
      else {
	raise_error(ERR_USAGE);
      }

    }
  }
  
  /* The last nfiles command arguments are the input file names */
  for (i = argc-nfiles; i < argc; i++) {
    config.team_fname[i-(argc-nfiles)] = (uint8_t *)malloc(strlen(argv[i])+1);
    if (!config.team_fname[i-(argc-nfiles)]) {
      raise_error(ERR_NOMEM);
    }
    strcpy(config.team_fname[i-(argc-nfiles)], argv[i]);
  }

  return;
}



/*****************************************************************************
 * Function: printConfig
 * Description: Prints the current configuration data. 
 *****************************************************************************/
void printConfig() {

  int8_t i;

  printf("\nCurrent configuration information:\n");
  printf("Debug level:           %d\n", config.debug_level);
  printf("Per-grid pixel width:  %d\n", config.screen_width);
  printf("Per-grid pixel height: %d\n", config.screen_height);
  
  for (i = 0; i < 4; i++) {
    printf("Team %d file name:      %s\n", i, config.team_fname[i]);
  }

  printf("Map file name:         %s\n", config.map_fname);

  printf("\n");

}





